import numpy as np
import torch

import memtorch
from memtorch.bh.crossbar.Program import naive_program
from memtorch.bh.nonideality.endurance_retention_models.conductance_drift import (
    model_conductance_drift,
)
from memtorch.bh.nonideality.endurance_retention_models.empirical_metal_oxide_RRAM import (
    model_endurance_retention,
)
from memtorch.map.Parameter import naive_map

valid_retention_models = [model_endurance_retention, model_conductance_drift]


def apply_retention_model(
    layer, time, retention_model=model_conductance_drift, **retention_model_kwargs
):
    """Method to apply an retention model to devices within a memristive layer.

    Parameters
    ----------
    layer : memtorch.mn
        A memrstive layer.
    time : float
        Retention time (s).
    retention_model : function
        Retention model to use.
    retention_model_kwargs : **kwargs
        Retention model keyword arguments.

    Returns
    -------
    memtorch.mn
        The patched memristive layer.
    """
    assert retention_model in valid_retention_models, "retention_model is not valid."
    return retention_model(layer=layer, x=time, **retention_model_kwargs)


if __name__ == "__main__":

    class MLP(torch.nn.Module):
        def __init__(self, n_inputs):
            super(MLP, self).__init__()
            self.layer = torch.nn.Linear(n_inputs, 10)
            self.activation = torch.nn.ReLU()

        def forward(self, x):
            x = self.layer(x)
            x = self.activation(x)
            return x

    model = MLP(n_inputs=100)
    reference_memristor = memtorch.bh.memristor.VTEAM
    m_model = patch_model(
        model,
        memristor_model=reference_memristor,
        memristor_model_params={},
        module_parameters_to_patch=[torch.nn.Linear],
        mapping_routine=naive_map,
        transistor=True,
        programming_routine=None,
        scheme=memtorch.bh.Scheme.DoubleColumn,
        tile_shape=(8, 8),
        max_input_voltage=0.3,
        ADC_resolution=int(8),
        ADC_overflow_rate=0.0,
        quant_method="linear",
    )
    m_model.tune_()
    p_model = memtorch.bh.nonideality.apply_nonidealities(
        m_model,
        non_idealities=[memtorch.bh.nonideality.NonIdeality.Retention],
        time=1e4,
        retention_model=memtorch.bh.nonideality.endurance_retention_models.model_conductance_drift,
        retention_model_kwargs={"initial_time": 0.0, "drift_coefficient": 0.1},
    )
    m_model.tune_()



import glob
import os

import torch
from setuptools import find_packages, setup
from torch.utils.cpp_extension import include_paths, library_paths

version = "1.1.6"
CUDA = True
CUDA_device_idx = 0  # Optional, ignored if CUDA is False


if CUDA:
    torch.cuda.set_device(torch.device("cuda", CUDA_device_idx))


def create_version_py(version, CUDA):
    file = open("memtorch/version.py", "w")
    if CUDA:
        version_string = "__version__ = '{}'".format(version)
    else:
        version_string = "__version__ = '{}-cpu'".format(version)

    file.write(version_string)
    file.close()


create_version_py(version, CUDA)
if CUDA:
    from torch.utils.cpp_extension import BuildExtension, CppExtension, CUDAExtension

    ext_modules = [
        CUDAExtension(
            name="memtorch_cuda_bindings",
            sources=glob.glob("memtorch/cu/*.cu")
            + glob.glob("memtorch/cu/*.cpp")
            + ["memtorch/cpp/gen_tiles.cpp"],
            library_dirs=["memtorch/submodules"],
            include_dirs=[
                os.path.join(os.getcwd(), relative_path)
                for relative_path in [
                    "memtorch/cu/",
                    "memtorch/submodules/eigen/",
                ]
            ],
            extra_compile_args=["-lineinfo", "-use_fast_math"],
        ),
        CppExtension(
            name="memtorch_bindings",
            sources=glob.glob("memtorch/cpp/*.cpp"),
            include_dirs=[
                os.path.join(os.getcwd(), relative_path)
                for relative_path in [
                    "memtorch/cpp/",
                    "memtorch/submodules/eigen/",
                ]
            ],
            extra_compile_args=["-O3"],
        ),
    ]
    name = "memtorch"
else:
    from torch.utils.cpp_extension import BuildExtension, CppExtension

    ext_modules = [
        CppExtension(
            name="memtorch_bindings",
            sources=glob.glob("memtorch/cpp/*.cpp"),
            include_dirs=[
                os.path.join(os.getcwd(), relative_path)
                for relative_path in [
                    "memtorch/cpp/",
                    "memtorch/submodules/eigen/",
                ]
            ],
            extra_compile_args=["-O3"],
        )
    ]
    name = "memtorch-cpu"

if __name__ == "__main__":
    setup(
        name=name,
        version=version,
        description="A Simulation Framework for Memristive Deep Learning Systems",
        long_description="A Simulation Framework for Memristive Deep Learning Systems which integrates directly with the well-known PyTorch Machine Learning (ML) library",
        url="https://github.com/coreylammie/MemTorch",
        license="GPL-3.0",
        author="Corey Lammie",
        author_email="coreylammie@jcu.edu.au",
        setup_requires=["ninja"],
        ext_modules=ext_modules,
        cmdclass={"build_ext": BuildExtension},
        packages=find_packages(),
        install_requires=[
            "numpy",
            "pandas",
            "scipy",
            "sklearn",
            "torch>=1.2.0",
            "torchvision",
            "matplotlib",
            "seaborn",
            "ipython",
            "lmfit",
        ],
        include_package_data=True,
        python_requires=">=3.6",
    )



import matplotlib
import numpy as np
import pytest
import torch

import memtorch


def get_subclasses(cls):
    for subclass in cls.__subclasses__():
        yield from get_subclasses(subclass)
        yield subclass


matplotlib.use("Agg")
memristor_models = list(get_subclasses(memtorch.bh.memristor.Memristor))


@pytest.mark.parametrize("model", memristor_models)
def test_model(model):
    model_instance = model()
    assert model_instance.g is not None
    assert model_instance.r_on is not None
    assert model_instance.r_off is not None
    assert model_instance.time_series_resolution is not None


@pytest.mark.parametrize("model", memristor_models)
@pytest.mark.filterwarnings("ignore::UserWarning")
def test_plot_hysteresis_loop(model):
    model_instance = model()
    voltage_signal, current_signal = model_instance.plot_hysteresis_loop(
        return_result=True
    )
    assert voltage_signal is not None and current_signal is not None
    assert len(voltage_signal) == len(current_signal)
    assert model_instance.plot_hysteresis_loop(return_result=False) is None
    assert model_instance.plot_bipolar_switching_behaviour(return_result=False) is None


@pytest.mark.parametrize("model", memristor_models)
def test_simulate(model):
    model_instance = model()
    voltage_signal = np.zeros((100, 1))
    assert len(model_instance.simulate(voltage_signal, return_current=True)) == 100
    assert model_instance.simulate(voltage_signal, return_current=False) is None



import inspect

import numpy as np
import torch

import memtorch


def apply_device_faults(layer, lrs_proportion, hrs_proportion, electroform_proportion):
    """Method to model device failure within a memristive layer.

    Parameters
    ----------
    layer : memtorch.mn
        A memrstive layer.
    lrs_proportion : float
        Proportion of devices which become stuck at a low resistance state.
    hrs_proportion : float
        Proportion of devices which become stuck at a high resistance state.
    electroform_proportion : float
        Proportion of devices which fail to electroform.

    Returns
    -------
    memtorch.mn
        The patched memristive layer.
    """
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")

    def apply_device_faults_to_crossbar(crossbar, lrs_proportion, hrs_proportion):
        crossbar_min = (
            torch.tensor(1 / (np.vectorize(lambda x: x.r_off)(crossbar.devices)))
            .view(-1)
            .to(device)
        )
        crossbar_max = (
            torch.tensor(1 / (np.vectorize(lambda x: x.r_on)(crossbar.devices)))
            .view(-1)
            .to(device)
        )
        crossbar_min_indices = np.random.choice(
            np.arange(torch.numel(crossbar_min)),
            replace=False,
            size=int(torch.numel(crossbar_min) * hrs_proportion),
        )
        crossbar_shape = crossbar.conductance_matrix.shape
        crossbar.conductance_matrix = crossbar.conductance_matrix.reshape(-1)
        for index in crossbar_min_indices:
            crossbar.conductance_matrix[index] = crossbar_min[index]

        crossbar_max_indices = np.random.choice(
            np.arange(torch.numel(crossbar_max)),
            replace=False,
            size=int(torch.numel(crossbar_max) * lrs_proportion),
        )
        for index in crossbar_max_indices:
            crossbar.conductance_matrix[index] = crossbar_max[index]

        crossbar.conductance_matrix = crossbar.conductance_matrix.view(crossbar_shape)
        return crossbar

    hrs_proportion = hrs_proportion + electroform_proportion
    for i in range(len(layer.crossbars)):
        layer.crossbars[i] = apply_device_faults_to_crossbar(
            layer.crossbars[i], lrs_proportion, hrs_proportion
        )

    return layer


def apply_cycle_variability(
    layer,
    distribution=torch.distributions.normal.Normal,
    min=0,
    max=float("Inf"),
    parallelize=False,
    r_off_kwargs={},
    r_on_kwargs={},
):
    """Method to apply cycle-to-cycle variability to a memristive layer.

    Parameters
    ----------
    layer : memtorch.mn
        A memrstive layer.
    distribution : torch.distributions
        torch distribution.
    min : float
        Minimum value to sample.
    max: float
        Maximum value to sample.
    parallelize : bool
        The operation is parallelized (True).
    r_off_kwargs : dict
        r_off kwargs.
    r_on_kwargs : dict
        r_on kwargs.
    """
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")

    def apply_cycle_variability_to_crossbar(
        crossbar,
        distribution=torch.distributions.normal.Normal,
        min=0,
        max=float("Inf"),
        parallelize=False,
        r_off_kwargs={},
        r_on_kwargs={},
    ):
        assert issubclass(
            distribution, torch.distributions.distribution.Distribution
        ), "Distribution is not in torch.distributions."
        for arg in inspect.signature(distribution).parameters.values():
            if (
                arg.name not in r_off_kwargs or arg.name not in r_on_kwargs
            ) and arg.name is not "validate_args":
                raise Exception(
                    "Argument %s is required for %s" % (arg.name, distribution)
                )

        shape = crossbar.conductance_matrix.shape
        r_off = distribution(**r_off_kwargs).sample(sample_shape=shape).clamp(min, max)
        r_on = distribution(**r_on_kwargs).sample(sample_shape=shape).clamp(min, max)
        if parallelize:

            def write_r_off(device, conductance):
                device.r_off = conductance

            def write_r_on(device, conductance):
                device.r_on = conductance

            np.frompyfunc(write_r_off, 2, 0)(crossbar.devices, r_off)
            np.frompyfunc(write_r_on, 2, 0)(crossbar.devices, r_on)
        else:
            if layer.tile_shape is not None:
                for i in range(0, layer.crossbars[0].devices.shape[0]):
                    for j in range(0, layer.crossbars[0].devices.shape[1]):
                        for k in range(0, layer.crossbars[0].devices.shape[2]):
                            crossbar.devices[i][j][k].r_off = r_off[i][j][k].item()
                            crossbar.devices[i][j][k].r_on = r_on[i][j][k].item()
            else:
                for i in range(0, crossbar.rows):
                    for j in range(0, crossbar.columns):
                        crossbar.devices[i][j].r_off = r_off[i][j].item()
                        crossbar.devices[i][j].r_on = r_on[i][j].item()

        crossbar.conductance_matrix = torch.max(
            torch.min(crossbar.conductance_matrix.clone().detach().cpu(), 1 / r_on),
            1 / r_off,
        ).to(device)
        crossbar.update(from_devices=False)
        return crossbar

    for i in range(len(layer.crossbars)):
        layer.crossbars[i] = apply_cycle_variability_to_crossbar(
            layer.crossbars[i],
            distribution,
            min=min,
            max=max,
            parallelize=parallelize,
            r_off_kwargs=r_off_kwargs,
            r_on_kwargs=r_on_kwargs,
        )

    return layer



import numpy as np
import torch
import torch.functional as F
import torch.nn as nn

import memtorch
from memtorch.utils import convert_range


def naive_map(weight, r_on, r_off, scheme, p_l=None):
    """Method to naively map network parameters to memristive device conductances, using two crossbars to represent both positive and negative weights.

    Parameters
    ----------
    weight : torch.Tensor
        Weight tensor to map.
    r_on : float
        Low resistance state.
    r_off : float
        High resistance state.
    scheme: memtorch.bh.crossbar.Scheme
        Weight representation scheme.
    p_l: float, optional
        If not None, the proportion of weights to retain.

    Returns
    -------
    torch.Tensor, torch.Tensor
        Positive and negative crossbar weights.
    """
    if p_l is not None:
        assert p_l >= 0 and p_l <= 1, "p_l must be None or between 0 and 1."
        weight_max = sorted(weight.abs().flatten().cpu().detach().numpy(), reverse=True)
        weight_max = weight_max[int(p_l * (weight.numel() - 1))]
        weight_min = weight_max / (r_off / r_on)
    else:
        weight_max = weight.abs().max()
        weight_min = 0

    if scheme == memtorch.bh.crossbar.Scheme.DoubleColumn:
        pos = weight.clone()
        neg = weight.clone() * -1
        pos[pos < 0] = 0
        neg[neg < 0] = 0
        pos = torch.clamp(pos, weight_min, weight_max)
        neg = torch.clamp(neg, weight_min, weight_max)
        pos = convert_range(pos, weight_min, weight_max, 1 / r_off, 1 / r_on)
        neg = convert_range(neg, weight_min, weight_max, 1 / r_off, 1 / r_on)
        return pos, neg
    elif scheme == memtorch.bh.crossbar.Scheme.SingleColumn:
        crossbar = weight.clone()
        crossbar = torch.clamp(crossbar, weight_min, weight_max)
        crossbar = convert_range(crossbar, weight_min, weight_max, 1 / r_off, 1 / r_on)
        return crossbar
    else:
        raise ("%s is not currently supported." % scheme)



import copy

import numpy as np
import torch

import memtorch
import memtorch_bindings

quant_methods = ["linear", "log"]


def quantize(
    tensor,
    quant,
    overflow_rate=0.0,
    quant_method=None,
    min=float("nan"),
    max=float("nan"),
    override_original=False,
):
    """Method to quantize a tensor.

    Parameters
    ----------
    tensor : torch.Tensor
        Input tensor.
    quant : int
        Bit width (if quant_method is not None) or the number of discrete quantization levels (if quant_method is None).
    overflow_rate : float, optional
        Overflow rate threshold for linear quantization.
    quant_method : str, optional
        Quantization method. Must be in quant_methods.
    min : float or tensor, optional
        Minimum value(s) to clip numbers to.
    max : float or tensor, optional
        Maximum value(s) to clip numbers to.
    override_original : bool, optional
        Whether to override the original tensor (True) or not (False).

    Returns
    -------
    torch.Tensor
        Quantized tensor.

    """
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    assert (
        overflow_rate >= 0 and overflow_rate <= 1
    ), "overflow_rate must be >= 0 and <= 1."
    assert (
        type(quant) == int and quant > 0
    ), "The bit width or number of discrete quantization levels must be a positive integer."
    if type(min) == int:
        min = float(min)
    if type(max) == int:
        max = float(max)
    if not override_original:
        tensor = copy.deepcopy(tensor)

    tensor = tensor.cpu()
    if tensor.dtype != torch.float32 and tensor.dtype != torch.float64:
        tensor = tensor.float()

    if quant_method is not None:
        assert quant_method in quant_methods, "quant_method is invalid."
        memtorch_bindings.quantize(
            tensor,
            bits=quant,
            overflow_rate=overflow_rate,
            quant_method=quant_methods.index(quant_method),
            min=min,
            max=max,
        )
    else:
        memtorch_bindings.quantize(tensor, n_quant_levels=quant, min=min, max=max)

    return tensor.to(device)



import numpy as np
import torch

import memtorch

if "cpu" not in memtorch.__version__:
    import memtorch_cuda_bindings

import memtorch_bindings


def solve_passive(
    conductance_matrix,
    V_WL,
    V_BL,
    R_source,
    R_line,
    n_input_batches=None,
    det_readout_currents=True,
    use_bindings=True,
    cuda_malloc_heap_size=None,
):
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    assert R_source != 0 or R_line != 0, "R_source or R_line must be non-zero."
    assert R_source >= 0 and R_line >= 0, "R_source and R_line must be >=0."
    if use_bindings:
        if n_input_batches is None:
            if "cpu" in memtorch.__version__:
                return memtorch_bindings.solve_passive(
                    conductance_matrix,
                    V_WL,
                    V_BL,
                    R_source,
                    R_line,
                    det_readout_currents=det_readout_currents,
                )
            else:
                if cuda_malloc_heap_size is None:
                    return memtorch_cuda_bindings.solve_passive(
                        conductance_matrix,
                        V_WL,
                        V_BL,
                        R_source,
                        R_line,
                        det_readout_currents=det_readout_currents,
                    )
                else:
                    return memtorch_cuda_bindings.solve_passive(
                        conductance_matrix,
                        V_WL,
                        V_BL,
                        R_source,
                        R_line,
                        det_readout_currents=det_readout_currents,
                        cuda_malloc_heap_size=cuda_malloc_heap_size,
                    )
        else:
            return memtorch_bindings.solve_passive(
                conductance_matrix.cpu(),
                V_WL.cpu(),
                V_BL.cpu(),
                R_source,
                R_line,
                n_input_batches=n_input_batches,
            ).to(device)
    else:
        m = conductance_matrix.shape[0]
        n = conductance_matrix.shape[1]
        indices = torch.zeros(2, 8 * m * n - 2 * m - 2 * n, device=device)
        values = torch.zeros(8 * m * n - 2 * m - 2 * n, device=device)
        mn_range = torch.arange(m * n)
        m_range = torch.arange(m)
        n_range = torch.arange(n)
        index = 0
        # A matrix
        for i in range(m):
            indices[0:2, index] = i * n
            if R_source == 0:
                values[index] = conductance_matrix[i, 0] + 1 / R_line
            elif R_line == 0:
                values[index] = conductance_matrix[i, 0] + 1 / R_source
            else:
                values[index] = conductance_matrix[i, 0] + 1 / R_source + 1 / R_line

            index += 1
            indices[0, index] = i * n + 1
            indices[1, index] = i * n
            if R_line == 0:
                values[index : index + 2] = 0
            else:
                values[index : index + 2] = -1 / R_line

            index += 1
            indices[0, index] = i * n
            indices[1, index] = i * n + 1
            index += 1
            indices[0:2, index] = i * n + (n - 1)
            if R_line == 0:
                values[index] = conductance_matrix[i, n - 1]
            else:
                values[index] = conductance_matrix[i, n - 1] + 1 / R_line

            index += 1
            for j in range(1, n - 1):
                indices[0:2, index] = i * n + j
                if R_line == 0:
                    values[index] = conductance_matrix[i, j]
                else:
                    values[index] = conductance_matrix[i, j] + 2 / R_line
                index += 1
                indices[0, index] = i * n + j + 1
                indices[1, index] = i * n + j
                if R_line == 0:
                    values[index : index + 2] = 0
                else:
                    values[index : index + 2] = -1 / R_line

                index += 1
                indices[0, index] = i * n + j
                indices[1, index] = i * n + j + 1
                index += 1

        # B matrix
        indices[0, index : index + (m * n)] = mn_range
        indices[1, index : index + (m * n)] = (
            indices[0, index : index + (m * n)] + m * n
        )
        values[index : index + (m * n)] = -conductance_matrix[
            n_range.repeat_interleave(m), n_range.repeat(m)
        ]
        index = index + (m * n)
        # C matrix
        indices[0, index : index + (m * n)] = mn_range + m * n
        del mn_range
        indices[1, index : index + (m * n)] = n * m_range.repeat(
            n
        ) + n_range.repeat_interleave(m)
        values[index : index + (m * n)] = conductance_matrix[
            m_range.repeat_interleave(n), n_range.repeat(m)
        ]
        index = index + (m * n)
        # D matrix
        for j in range(n):
            indices[0, index] = m * n + (j * m)
            indices[1, index] = m * n + j
            if R_line == 0:
                values[index] = -conductance_matrix[0, j]
            else:
                values[index] = -1 / R_line - conductance_matrix[0, j]

            index += 1
            indices[0, index] = m * n + (j * m)
            indices[1, index] = m * n + j + n
            if R_line == 0:
                values[index : index + 2] = 0
            else:
                values[index : index + 2] = 1 / R_line

            index += 1
            indices[0, index : index + 2] = m * n + (j * m) + m - 1
            indices[1, index] = m * n + (n * (m - 2)) + j
            index += 1
            indices[1, index] = m * n + (n * (m - 1)) + j
            if R_source == 0:
                values[index] = -conductance_matrix[m - 1, j] - 1 / R_line
            elif R_line == 0:
                values[index] = -1 / R_source - conductance_matrix[m - 1, j]
            else:
                values[index] = (
                    -1 / R_source - conductance_matrix[m - 1, j] - 1 / R_line
                )

            index += 1
            indices[0, index : index + 3 * (m - 2)] = (
                m * n + (j * m) + m_range[1:-1].repeat_interleave(3)
            )
            for i in range(1, m - 1):
                indices[1, index] = m * n + (n * (i - 1)) + j
                if R_line == 0:
                    values[index : index + 2] = 0
                else:
                    values[index : index + 2] = 1 / R_line

                index += 1
                indices[1, index] = m * n + (n * (i + 1)) + j
                index += 1
                indices[1, index] = m * n + (n * i) + j
                if R_line == 0:
                    values[index] = -conductance_matrix[i, j]
                else:
                    values[index] = -conductance_matrix[i, j] - 2 / R_line

                index += 1

        if n_input_batches is None:
            E_matrix = torch.zeros(2 * m * n, device=device)
            if R_source == 0:
                E_matrix[m_range * n] = V_WL.to(device)  # E_W values
                E_matrix[m * n + (n_range + 1) * m - 1] = -V_BL.to(device)  # E_B values
            else:
                # E_W values
                E_matrix[m_range * n] = V_WL.to(device) / R_source
                E_matrix[m * n + (n_range + 1) * m - 1] = (
                    -V_BL.to(device) / R_source
                )  # E_B values

            V = torch.linalg.solve(
                torch.sparse_coo_tensor(
                    indices, values, (2 * m * n, 2 * m * n), device=device
                ).to_dense(),
                E_matrix,
            )
            V_applied = torch.zeros((m, n), device=device)
            for i in m_range:
                V_applied[i, n_range] = V[n * i + n_range] - V[m * n + n * i + n_range]
            if not det_readout_currents:
                return V_applied
            else:
                return torch.sum(torch.mul(V_applied, conductance_matrix.to(device)), 0)
        else:
            out = torch.zeros(n_input_batches, n, device=device)
            for i in range(n_input_batches):
                E_matrix = torch.zeros(2 * m * n, device=device)
                if R_source == 0:
                    E_matrix[m_range * n] = V_WL[i, :].to(device)  # E_W values
                    E_matrix[m * n + (n_range + 1) * m - 1] = -V_BL[i, :].to(
                        device
                    )  # E_B values
                else:
                    E_matrix[m_range * n] = (
                        V_WL[i, :].to(device) / R_source
                    )  # E_W values
                    E_matrix[m * n + (n_range + 1) * m - 1] = (
                        -V_BL[i, :].to(device) / R_source
                    )  # E_B values

                V = torch.linalg.solve(
                    torch.sparse_coo_tensor(
                        indices, values, (2 * m * n, 2 * m * n), device=device
                    ).to_dense(),
                    E_matrix,
                )
                V_applied = torch.zeros((m, n), device=device)
                for j in m_range:
                    V_applied[j, n_range] = (
                        V[n * j + n_range] - V[m * n + n * j + n_range]
                    )

                out[i, :] = torch.sum(
                    torch.mul(V_applied, conductance_matrix.to(device)), 0
                )

            return out



import copy
import math

import numpy as np
import pytest
import torch

import memtorch
from memtorch.bh.nonideality.DeviceFaults import apply_cycle_variability
from memtorch.bh.nonideality.endurance_retention_models.conductance_drift import (
    model_conductance_drift,
)
from memtorch.bh.nonideality.endurance_retention_models.empirical_metal_oxide_RRAM import (
    model_endurance_retention,
)
from memtorch.bh.nonideality.NonIdeality import apply_nonidealities
from memtorch.mn.Module import supported_module_parameters


@pytest.mark.parametrize("tile_shape", [None, (128, 128), (10, 20)])
@pytest.mark.parametrize("quant_method", memtorch.bh.Quantize.quant_methods + [None])
def test_device_faults(debug_patched_networks, tile_shape, quant_method):
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    patched_networks = debug_patched_networks(tile_shape, quant_method)
    for patched_network in patched_networks:
        patched_network_lrs = apply_nonidealities(
            copy.deepcopy(patched_network),
            non_idealities=[memtorch.bh.nonideality.NonIdeality.DeviceFaults],
            lrs_proportion=0.5,
            hrs_proportion=0,
            electroform_proportion=0,
        )
        patched_tensor_lrs = (
            patched_network_lrs.layer.crossbars[0].conductance_matrix.float().to(device)
        )
        lrs = (
            torch.tensor(
                1
                / np.vectorize(lambda x: x.r_on)(
                    patched_network_lrs.layer.crossbars[0].devices
                )
            )
            .float()
            .to(device)
        )
        lrs_percentage = (
            sum(torch.isclose(patched_tensor_lrs, lrs).view(-1)).item()
            / patched_tensor_lrs.numel()
        )
        patched_network_hrs = apply_nonidealities(
            copy.deepcopy(patched_network),
            non_idealities=[memtorch.bh.nonideality.NonIdeality.DeviceFaults],
            lrs_proportion=0,
            hrs_proportion=0.25,
            electroform_proportion=0.25,
        )
        patched_tensor_hrs = (
            patched_network_hrs.layer.crossbars[0].conductance_matrix.float().to(device)
        )
        hrs = (
            torch.tensor(
                1
                / np.vectorize(lambda x: x.r_off)(
                    patched_network_hrs.layer.crossbars[0].devices
                )
            )
            .float()
            .to(device)
        )
        hrs_percentage = (
            sum(torch.isclose(patched_tensor_hrs, hrs).view(-1)).item()
            / patched_tensor_hrs.numel()
        )
        assert (
            lrs_percentage >= 0.25 and hrs_percentage >= 0.25
        )  # To account for some degree of stochasticity


@pytest.mark.parametrize("tile_shape", [None, (128, 128), (10, 20)])
@pytest.mark.parametrize("quant_method", memtorch.bh.Quantize.quant_methods + [None])
def test_finite_conductance_states(
    debug_patched_networks, tile_shape, quant_method, conductance_states=5
):
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    patched_networks = debug_patched_networks(tile_shape, quant_method)
    for patched_network in patched_networks:
        patched_network_finite_states = apply_nonidealities(
            copy.deepcopy(patched_network),
            non_idealities=[
                memtorch.bh.nonideality.NonIdeality.FiniteConductanceStates
            ],
            conductance_states=5,
        )
        conductance_matrix = patched_network.layer.crossbars[0].conductance_matrix
        quantized_conductance_matrix = patched_network_finite_states.layer.crossbars[
            0
        ].conductance_matrix
        quantized_conductance_matrix_unique = quantized_conductance_matrix.unique()
        valid_values = torch.linspace(
            patched_network.layer.crossbars[0].conductance_matrix.min(),
            patched_network.layer.crossbars[0].conductance_matrix.max(),
            conductance_states,
        ).float()
        assert any(
            [
                bool(val)
                for val in [
                    torch.isclose(
                        quantized_conductance_matrix_unique, valid_value
                    ).any()
                    for valid_value in valid_values
                ]
            ]
        )
        assert conductance_matrix.shape == quantized_conductance_matrix.shape


@pytest.mark.parametrize("tile_shape", [None, (128, 128), (10, 20)])
@pytest.mark.parametrize("parallelize", [True, False])
@pytest.mark.parametrize("quant_method", memtorch.bh.Quantize.quant_methods + [None])
def test_cycle_variability(
    debug_patched_networks, tile_shape, parallelize, quant_method, std=10
):
    patched_networks = debug_patched_networks(tile_shape, quant_method)
    for patched_network in patched_networks:
        for i, (name, m) in enumerate(list(patched_network.named_modules())):
            if type(m) in supported_module_parameters.values():
                if "cpu" not in memtorch.__version__ and len(name.split(".")) > 1:
                    name = name.split(".")[1]

                if hasattr(patched_network, "module"):
                    with pytest.raises(Exception):
                        setattr(
                            patched_network.module,
                            name,
                            apply_cycle_variability(
                                m,
                                parallelize=parallelize,
                                r_off_kwargs={"invalid_arg": None},
                                r_on_kwargs={"invalid_arg": None},
                            ),
                        )

                    setattr(
                        patched_network.module,
                        name,
                        apply_cycle_variability(
                            m,
                            parallelize=parallelize,
                            r_off_kwargs={
                                "loc": m.crossbars[0].r_off_mean,
                                "scale": std * 2,
                            },
                            r_on_kwargs={"loc": m.crossbars[0].r_on_mean, "scale": std},
                        ),
                    )
                else:
                    with pytest.raises(Exception):
                        setattr(
                            patched_network,
                            name,
                            apply_cycle_variability(
                                m,
                                parallelize=parallelize,
                                r_off_kwargs={"invalid_arg": None},
                                r_on_kwargs={"invalid_arg": None},
                            ),
                        )

                    setattr(
                        patched_network,
                        name,
                        apply_cycle_variability(
                            m,
                            parallelize=parallelize,
                            r_off_kwargs={
                                "loc": m.crossbars[0].r_off_mean,
                                "scale": std * 2,
                            },
                            r_on_kwargs={"loc": m.crossbars[0].r_on_mean, "scale": std},
                        ),
                    )


@pytest.mark.parametrize("tile_shape", [None, (128, 128), (10, 20)])
@pytest.mark.parametrize("quant_method", memtorch.bh.Quantize.quant_methods + [None])
def test_non_linear(debug_patched_networks, tile_shape, quant_method):
    patched_networks = debug_patched_networks(tile_shape, quant_method)
    for patched_network in patched_networks:
        patched_network_non_linear = apply_nonidealities(
            copy.deepcopy(patched_network),
            non_idealities=[memtorch.bh.nonideality.NonIdeality.NonLinear],
            sweep_duration=2,
            sweep_voltage_signal_amplitude=1,
            sweep_voltage_signal_frequency=0.5,
        )
        patched_network_non_linear.tune_(
            tune_kwargs={
                "<class 'memtorch.mn.Conv1d.Conv1d'>": {
                    "input_batch_size": 1,
                    "input_shape": 2,
                },
                "<class 'memtorch.mn.Conv2d.Conv2d'>": {
                    "input_batch_size": 1,
                    "input_shape": 2,
                },
                "<class 'memtorch.mn.Conv3d.Conv3d'>": {
                    "input_batch_size": 1,
                    "input_shape": 2,
                },
                "<class 'memtorch.mn.Linear.Linear'>": {"input_shape": 2},
            }
        )
        patched_network_non_linear = apply_nonidealities(
            copy.deepcopy(patched_network),
            non_idealities=[memtorch.bh.nonideality.NonIdeality.NonLinear],
            simulate=True,
        )
        patched_network_non_linear.tune_(
            tune_kwargs={
                "<class 'memtorch.mn.Conv1d.Conv1d'>": {
                    "input_batch_size": 1,
                    "input_shape": 2,
                },
                "<class 'memtorch.mn.Conv2d.Conv2d'>": {
                    "input_batch_size": 1,
                    "input_shape": 2,
                },
                "<class 'memtorch.mn.Conv3d.Conv3d'>": {
                    "input_batch_size": 1,
                    "input_shape": 2,
                },
                "<class 'memtorch.mn.Linear.Linear'>": {"input_shape": 2},
            }
        )


@pytest.mark.parametrize("tile_shape", [None, (128, 128), (10, 20)])
@pytest.mark.parametrize(
    "operation_mode",
    [
        memtorch.bh.nonideality.endurance_retention_models.OperationMode.sudden,
        memtorch.bh.nonideality.endurance_retention_models.OperationMode.gradual,
    ],
)
@pytest.mark.parametrize("temperature", [350, None])
def test_model_endurance_retention_retention(
    debug_patched_networks,
    tile_shape,
    operation_mode,
    temperature,
    time=1e4,
    p_lrs=[1, 0, 0, 0],
    stable_resistance_lrs=100,
    p_hrs=[1, 0, 0, 0],
    stable_resistance_hrs=1000,
    cell_size=None,
):
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    patched_networks = debug_patched_networks(tile_shape, None)
    for patched_network in patched_networks:
        patched_network = apply_nonidealities(
            copy.deepcopy(patched_network),
            non_idealities=[memtorch.bh.nonideality.NonIdeality.Retention],
            time=float(time),
            retention_model=memtorch.bh.nonideality.endurance_retention_models.model_endurance_retention,
            retention_model_kwargs={
                "operation_mode": operation_mode,
                "p_lrs": p_lrs,
                "stable_resistance_lrs": stable_resistance_lrs,
                "p_hrs": p_hrs,
                "stable_resistance_hrs": stable_resistance_hrs,
                "cell_size": cell_size,
                "temperature": temperature,
            },
        )


@pytest.mark.parametrize("tile_shape", [None, (128, 128), (10, 20)])
@pytest.mark.parametrize(
    "operation_mode",
    [
        memtorch.bh.nonideality.endurance_retention_models.OperationMode.sudden,
        memtorch.bh.nonideality.endurance_retention_models.OperationMode.gradual,
    ],
)
@pytest.mark.parametrize("temperature", [350, None])
def test_model_endurance_retention_endurance(
    debug_patched_networks,
    tile_shape,
    operation_mode,
    temperature,
    x=1e4,
    p_lrs=[1, 0, 0, 0],
    stable_resistance_lrs=100,
    p_hrs=[1, 0, 0, 0],
    stable_resistance_hrs=1000,
    cell_size=None,
):
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    patched_networks = debug_patched_networks(tile_shape, None)
    for patched_network in patched_networks:
        patched_network = apply_nonidealities(
            copy.deepcopy(patched_network),
            non_idealities=[memtorch.bh.nonideality.NonIdeality.Endurance],
            x=float(x),
            endurance_model=memtorch.bh.nonideality.endurance_retention_models.model_endurance_retention,
            endurance_model_kwargs={
                "operation_mode": operation_mode,
                "p_lrs": p_lrs,
                "stable_resistance_lrs": stable_resistance_lrs,
                "p_hrs": p_hrs,
                "stable_resistance_hrs": stable_resistance_hrs,
                "cell_size": cell_size,
                "temperature": temperature,
            },
        )


@pytest.mark.parametrize("v_stop", [0.25, 0.5])
@pytest.mark.parametrize("v_stop_optimal", [0.4, 0.6])
@pytest.mark.parametrize("cell_size", [10, None])
def test_scale_p_0(
    v_stop, v_stop_optimal, cell_size, p_0=1, p_1=1, v_stop_min=0, v_stop_max=1
):
    p_0 = memtorch.bh.nonideality.endurance_retention_models.scale_p_0(
        p_0, p_1, v_stop, v_stop_min, v_stop_max, v_stop_optimal, cell_size
    )


@pytest.mark.parametrize("time", [1e4, 1e6])
@pytest.mark.parametrize("drift_coefficient", [0.1, 0.4])
def test_model_conductance_drift(
    debug_patched_networks,
    time,
    drift_coefficient,
    tile_shape=(128, 128),
    initial_time=1e-12,
):
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    patched_networks = debug_patched_networks(tile_shape, None)
    for patched_network in patched_networks:
        patched_network = apply_nonidealities(
            copy.deepcopy(patched_network),
            non_idealities=[memtorch.bh.nonideality.NonIdeality.Retention],
            time=float(time),
            retention_model=memtorch.bh.nonideality.endurance_retention_models.model_conductance_drift,
            retention_model_kwargs={
                "initial_time": initial_time,
                "drift_coefficient": drift_coefficient,
            },
        )
